import pyspark.sql.functions as f
from ast import operator
from collections import namedtuple
from datetime import datetime
from dateutil.relativedelta import relativedelta
from operator import eq, gt, lt, ge, le, ne
from typing import Union, List, Dict


class Agile:
    def avg(
        self,
        df,
        ref: Union[str, list],
        iteration_path: str,
        new: str,
        literal_filter: List[str] = None,
        between_date: Dict[str, str] = None,
        group_by: List[str] = None,
        **filters,
    ):
        """
        After filtering a dataframe performs the arithmetic average of the column as a reference.
        """

        agg_column, df_callable = self.__filters(
            df, ref, literal_filter, between_date, **filters
        )
        df_result = (
            df_callable.groupby(iteration_path if not group_by else group_by)
            .agg(f.avg(agg_column).alias(new))
            .orderBy(new)
        )
        return Detail(df_result, df_callable)

    def count(
        self,
        df,
        ref: Union[str, list],
        iteration_path: str,
        new: str,
        literal_filter: List[str] = None,
        between_date: Dict[str, str] = None,
        group_by: List[str] = None,
        **filters,
    ):
        """
        After filtering a dataframe perform the column count as a reference.
        """
        agg_column, df_callable = self.__filters(
            df, ref, literal_filter, between_date, **filters
        )
        df_result = (
            df_callable.groupby(iteration_path if not group_by else group_by)
            .agg(f.countDistinct(agg_column).alias(new))
            .orderBy(new)
        )
        return Detail(df_result, df_callable)

    def custom(
        self,
        df_left,
        df_right,
        left: str,
        right: str,
        how: str,
        op: operator,
        left_ref: str,
        right_ref: str,
        new: str,
    ):
        """
        Agile.custom receives two PySpark dataframes and the information needed to apply the join and the signature of a Python operator that will do the operation between the two columns.
        Supported Operators: is_, is_not, add, and_, floordiv, mod, mul, pow, sub and ceil(pyspark).
        """
        temp_join = df_left.join(df_right, df_left[left] == df_right[right], how)
        df_temp = self.__interpreter_operator(
            temp_join, op, left_ref, right_ref, new
        ).drop(df_right[right])
        df_result = df_temp.select(left, new)
        return Detail(df_result, df_temp)

    def max(
        self,
        df,
        ref: Union[str, list],
        iteration_path: str,
        new: str,
        literal_filter: List[str] = None,
        between_date: Dict[str, str] = None,
        group_by: List[str] = None,
        **filters,
    ):
        """
        After filtering a dataframe perform the column max as a reference.
        """
        agg_column, df_callable = self.__filters(
            df, ref, literal_filter, between_date, **filters
        )
        df_result = (
            df_callable.groupby(iteration_path if not group_by else group_by)
            .agg(f.max(agg_column).alias(new))
            .orderBy(new)
        )
        return Detail(df_result, df_callable)

    def min(
        self,
        df,
        ref: Union[str, list],
        iteration_path: str,
        new: str,
        literal_filter: List[str] = None,
        between_date: Dict[str, str] = None,
        group_by: List[str] = None,
        **filters,
    ):
        """
        After filtering a dataframe perform the column min as a reference.
        """
        agg_column, df_callable = self.__filters(
            df, ref, literal_filter, between_date, **filters
        )
        df_result = (
            df_callable.groupby(iteration_path if not group_by else group_by)
            .agg(f.min(agg_column).alias(new))
            .orderBy(new)
        )
        return Detail(df_result, df_callable)

    def multiple_join(self, dfs: list, on: List[str], how: str = "left"):
        """
        Receives a list of dataframes and merges using the same column name between them.
        """
        previous = dfs[0]
        for df in dfs[1:]:
            merge = previous.join(df, on, how)
            previous = merge
        return merge

    def sum(
        self,
        df,
        ref: Union[str, list],
        iteration_path: str,
        new: str,
        literal_filter: List[str] = None,
        between_date: Dict[str, str] = None,
        group_by: List[str] = None,
        **filters,
    ):
        """
        After filtering a dataframe, perform the sum of the column as a reference.
        """
        agg_column, df_callable = self.__filters(
            df, ref, literal_filter, between_date, **filters
        )
        df_result = (
            df_callable.groupby(iteration_path if not group_by else group_by)
            .agg(f.sum(agg_column).alias(new))
            .orderBy(new)
        )
        return Detail(df_result, df_callable)

    def __filters(
        self,
        df,
        ref: Union[str, list],
        literal_filter: List[str] = None,
        between_date: Dict[str, str] = None,
        **filters,
    ):
        df_filter = self.__wrapper_filter(df, **filters)
        filter_between = self.__wrapper_between(between_date, df_filter)
        df_days = (
            self.__interpreter_literal_filter(filter_between, literal_filter)
            if literal_filter
            else filter_between
        )

        func, col1, col2 = ref if isinstance(ref, list) else [None, None, None]
        df_callable = (
            df_days.withColumn("temp", func(df_days[col1], df_days[col2]))
            if not isinstance(ref, str)
            else df_days
        )
        agg_column = ref if isinstance(ref, str) else "temp"
        return agg_column, df_callable

    def __interpreter_operator(self, temp_join, op: operator, left_ref, right_ref, new):
        accept = (
            "is_",
            "is_not",
            "add",
            "and_",
            "truediv",
            "floordiv",
            "mod",
            "mul",
            "pow",
            "sub",
            "ceil",
        )
        if op in accept:
            operators = {
                "is_": temp_join.withColumn(
                    new, temp_join[left_ref] == temp_join[right_ref]
                )
                if op == "is_"
                else None,
                "is_not": temp_join.withColumn(
                    new, temp_join[left_ref] != temp_join[right_ref]
                )
                if op == "is_not"
                else None,
                "add": temp_join.withColumn(
                    new, temp_join[left_ref] + temp_join[right_ref]
                )
                if op == "add"
                else None,
                "and_": temp_join.withColumn(
                    new, temp_join[left_ref] & temp_join[right_ref]
                )
                if op == "and_"
                else None,
                "truediv": temp_join.withColumn(
                    new, temp_join[left_ref] / temp_join[right_ref]
                )
                if op == "truediv"
                else None,
                "floordiv": temp_join.withColumn(
                    new, f.floor(temp_join[left_ref] / temp_join[right_ref])
                )
                if op == "floordiv"
                else None,
                "mod": temp_join.withColumn(
                    new, temp_join[left_ref] % temp_join[right_ref]
                )
                if op == "mod"
                else None,
                "mul": temp_join.withColumn(
                    new, temp_join[left_ref] * temp_join[right_ref]
                )
                if op == "mul"
                else None,
                "pow": temp_join.withColumn(
                    new, temp_join[left_ref] ** temp_join[right_ref]
                )
                if op == "pow"
                else None,
                "sub": temp_join.withColumn(
                    new, temp_join[left_ref] - temp_join[right_ref]
                )
                if op == "sub"
                else None,
                "ceil": temp_join.withColumn(
                    new, f.ceil(temp_join[left_ref] / temp_join[right_ref])
                )
                if op == "ceil"
                else None,
            }
            return operators.get(op)
        else:
            raise Exception(f"Operator {op} is not supported.")

    def __interpreter_literal_filter(self, df, expression: List[str]):
        accept = ("<", ">", ">=", "==", "<=", "!=")
        if expression:
            for exp in expression:
                expression_split = exp.split()
                col1, col2 = expression_split[0], expression_split[-1]
                validate = self.__validate_operator_dates(df, col1, col2)
                sig = [sig for sig in accept if sig in expression_split][0]
                operators = {
                    "<": df.filter(lt(df[col1], df[col2]))
                    if not validate
                    else df.filter(
                        lt(df[col1], (datetime.now() - relativedelta(days=int(col2))))
                    ),
                    ">": df.filter(gt(df[col1], df[col2]))
                    if not validate
                    else df.filter(
                        gt(df[col1], (datetime.now() - relativedelta(days=int(col2))))
                    ),
                    ">=": df.filter(ge(df[col1], df[col2]))
                    if not validate
                    else df.filter(
                        ge(df[col1], (datetime.now() - relativedelta(days=int(col2))))
                    ),
                    "==": df.filter(eq(df[col1], df[col2]))
                    if not validate
                    else df.filter(
                        eq(df[col1], (datetime.now() - relativedelta(days=int(col2))))
                    ),
                    "<=": df.filter(le(df[col1], df[col2]))
                    if not validate
                    else df.filter(
                        le(df[col1], (datetime.now() - relativedelta(days=int(col2))))
                    ),
                    "!=": df.filter(ne(df[col1], df[col2]))
                    if not validate
                    else df.filter(
                        ne(df[col1], (datetime.now() - relativedelta(days=int(col2))))
                    ),
                }
                df = operators.get(sig)
            return df

    def __validate_operator_dates(self, df, col1: str, col2: str):
        df_types = df.dtypes  # [('AreaPath', 'string'), ('CreatedDate', 'date'),...]
        is_numeric = True if col2.isnumeric() else False
        eq_col = [True for col in df_types if col[0] == col1][0]
        validate = True if is_numeric and eq_col else False
        return validate

    def __wrapper_between(self, between_date, df):
        be_created, be_closed = (
            self.__between_date(between_date) if between_date else [None, None]
        )
        filter_between = (
            df.where(
                (df[be_created.name] >= be_created.date)
                & (df[be_closed.name] <= be_closed.date)
            )
            if between_date
            else df
        )
        return filter_between

    @staticmethod
    def __isdifferent(items: List[str]):
        verify = [True if item[0:2] == "<>" else False for item in items]
        return verify

    @staticmethod
    def __clear_value(items: List[str]):
        return [item.replace("<>", "") for item in items]

    def __wrapper_filter(self, df, **filters):
        if filters:
            for key, value in filters["filters"].items():
                if isinstance(value, list):
                    hasdiff = self.__isdifferent(value)
                    if all(hasdiff):
                        df = df.where(~df[key].isin(self.__clear_value(value)))
                    elif True in hasdiff and False in hasdiff:
                        raise Exception(
                            "<> can only be used for all elements in the list."
                        )
                    else:
                        df = df.where(df[key].isin(value))
                else:
                    hasdiff = self.__isdifferent([value])
                    value = value.replace("<>", "")
                    if all(hasdiff):
                        df = df.filter(df[key] != value)
                    else:
                        if value == "null":
                            df = df.filter(df[key].isNull())
                        else:
                            df = df.filter(df[key] == value)
        return df

    @staticmethod
    def __between_date(between_date):
        if between_date:
            Bedate = namedtuple("BeDate", ["name", "date"])
            between = [item for item in between_date.items()]
            be_created = Bedate(between[0][0], f.lit(between[0][1]))
            be_closed = Bedate(between[1][0], f.lit(between[1][1]))
            return be_created, be_closed
        return [None, None]


class Detail:
    def __init__(self, df, detail):
        self.__dataframe = df
        self.__detail = detail

    @property
    def detail(self):
        """
        Returns the dataframe version before aggregation.
        """
        return self.__detail

    @property
    def df(self):
        """
        Returns the already aggregated dataframe.
        """
        return self.__dataframe



from datetime import datetime, date
from typing import List
from pyspark.sql.types import (
    BooleanType,
    DateType,
    DoubleType,
    IntegerType,
    StringType,
    StructField,
)


class Process:
    @classmethod
    def columns(cls, data: dict, fields_column: dict, fields_display: dict) -> dict:
        """
        Receives a dictionary with the data of a work item and returns the data with the selected fields.
        """
        fields = [
            {column.name: data[tag]["displayName"]}
            if tag in fields_display and tag in data
            else ({column.name: data[tag]} if tag in data else {column.name: None})
            for tag, column in fields_column.items()
        ]

        date_type_fields = [
            schema.name
            for name, schema in fields_column.items()
            if schema.dataType == DateType()
        ]
        schema_convert_date = [
            {name: cls.datetime_to_date(schema)}
            if name in date_type_fields and schema != None
            else {name: schema}
            for item in fields
            for name, schema in item.items()
        ]
        item = {}
        for value in schema_convert_date:
            for column, tag in value.items():
                item[column] = tag
        return item

    @classmethod
    def teams(cls, response: dict) -> List[dict]:
        """
        Receives data from all squads and returns `name and id`.
        """
        comprehension = [
            {
                "Squad": values["name"],
                "Id": values["id"],
            }
            for values in response["value"]
        ]
        return comprehension

    @classmethod
    def team_backlog(cls, response: dict) -> List[int]:
        """
        Receives a dictionary and returns a list of ids.
        """
        ids = []
        for list_value in response.values():
            for value in list_value:
                id = str(value["target"]["id"])
                ids.append(id)
        return ids

    @classmethod
    def team_interation(cls, response: dict, squad: str) -> List[dict]:
        """
        Receives a dictionary and returns a list of iteration information.
        """
        comprehension = [
            {
                "Iteration_Id": values["id"],
                "Iteration_Name": values["name"],
                "Iteration_Path": values["path"],
                "Iteration_Start_Date": cls.datetime_to_date(
                    values["attributes"]["startDate"]
                )
                if values["attributes"]["startDate"]
                else None,
                "Iteration_End_Date": cls.datetime_to_date(
                    values["attributes"]["finishDate"]
                )
                if values["attributes"]["finishDate"]
                else None,
                "Iteration_State": values["attributes"]["timeFrame"],
                "Squad": squad,
            }
            for values in response["value"]
        ]
        return comprehension

    @classmethod
    def members(cls, response: dict, squad: str) -> List[dict]:
        """
        Receives a dictionary and the name of a squad and returns a list of all squad members.
        """
        members = []
        for value in response["value"]:
            info = value["identity"]
            member = {"Name": info["displayName"], "Id": info["id"], "squad": squad}
            members.append(member)
        return members

    @staticmethod
    def datetime_to_date(str_datetime: str) -> date:
        """
        Receives a datetime in string and turns it into a date object.
        """
        date = str_datetime[: str_datetime.find("T")]
        return datetime.fromisoformat(date).date()

    @classmethod
    def tags(cls, response: dict) -> List[dict]:
        """
        Receives a dictionary and returns a list comprehension of all tags and their ids.
        """
        return [{"Id": item["id"], "Name": item["name"]} for item in response["value"]]

    @classmethod
    def all_process(cls, response: dict) -> List[dict]:
        """
        Receives a dictionary and returns a list of dict with process id and value.
        [{'id':...,'name': Basic}]
        """
        return [
            {"id": values["typeId"], "name": values["name"]}
            for values in response["value"]
        ]

    @classmethod
    def all_work_items(cls, response: dict) -> List[dict]:
        """
        Receives a dictionary and returns only the item id.
        """
        return [values["id"] for values in response["value"]]

    @classmethod
    def clean_fields(cls, response: dict) -> List[dict]:
        """
        Receives a dictionary with the process fields and returns only the referenceName, name and type.
        [ {'referenceName':'System.Tag', 'name':'Tag', 'type': 'string'},... ]
        """
        return [
            {
                "referenceName": values["referenceName"],
                "name": values["name"],
                "type": values["type"],
            }
            for values in response["value"]
        ]

    @classmethod
    def spark_equivalent(cls, fields: list) -> List[dict]:
        schema = {}
        spark_types = {
            "string": StringType(),
            "integer": IntegerType(),
            "double": DoubleType(),
            "boolean": BooleanType(),
            "html": StringType(),
            "dateTime": DateType(),
        }
        for item in fields:
            ref_name = item["referenceName"]
            name = item["referenceName"].split(".")[-1]
            type_field = spark_types.get(item["type"])
            field = StructField(name, type_field)
            schema[ref_name] = field
        return schema



from __future__ import annotations
from base64 import b64encode
from pyspark.sql.types import DateType, IntegerType, StringType, StructField, StructType
from pyspark.pandas.frame import DataFrame
from pyspark.sql import SparkSession
from typing import List
from requests import get, post
from copy import deepcopy
from array import array
from AzureDevopsPySpark.endpoints import *
from AzureDevopsPySpark.process import *
from AzureDevopsPySpark.schemas import *


class Azure:
    def __init__(self, organization: str, project: str, token: str):
        self._organization = organization
        self._project = project
        self._token = b64encode(f":{token}".encode()).decode()
        self.headers = {
            "Content-Type": "application/json",
            "Authorization": f"Basic {self._token}",
            "Proxy-Authorization": f"Basic {self._token}",
        }
        self._process = None
        self._columns = {
            "System.AreaPath": StructField("AreaPath", StringType(), True),
            "System.IterationPath": StructField("IterationPath", StringType(), True),
            "System.WorkItemType": StructField("WorkItemType", StringType(), True),
            "System.Id": StructField("Id", IntegerType(), True),
            "System.Parent": StructField("Parent", IntegerType(), True),
            "System.State": StructField("State", StringType(), True),
            "System.AssignedTo": StructField("AssignedTo", StringType(), True),
            "System.CreatedDate": StructField("CreatedDate", DateType(), False),
            "System.ChangedDate": StructField("ChangedDate", DateType(), True),
            "System.Title": StructField("Title", StringType(), True),
            "System.Tags": StructField("Tags", StringType(), True),
            "System.BoardColumn": StructField("BoardColumn", StringType(), True),
            "System.BoardColumnDone": StructField(
                "BoardColumnDone", StringType(), True
            ),
            "System.ChangedBy": StructField("ChangedBy", StringType(), True),
            "Microsoft.VSTS.Common.ActivatedDate": StructField(
                "ActivatedDate", DateType(), True
            ),
        }
        self._fields_displayname = [
            "System.AssignedTo",
            "System.ChangedBy",
            "Microsoft.VSTS.Common.ActivatedBy",
            "Microsoft.VSTS.Common.ClosedBy",
        ]
        self.spark = SparkSession.builder.appName("AzureDevopsSpark").getOrCreate()
        self.__validade_status()

    def __repr__(self):
        return f"Organization: {self._organization}\nProject: {self._project}\nToken base64: {self._token}\n{self.headers}"

    def filter_columns(self, only: List[str]) -> List[str, StructField]:
        """
        `Mapped columns that are not in the list passed as an argument will be excluded.`

        Ex:
            filter_columns( ['System.AreaPath', 'System.Id'] ) -> Only System.AreaPath and System.Id will be returned.
        """
        backup_keys = deepcopy(list(self._columns.keys()))
        return [self._columns.pop(item) for item in backup_keys if item not in only]

    def backlog(self) -> DataFrame:
        """
        Returns all backlog work items within a project.
        """
        items = []
        squads = [item["Squad"] for item in self.teams().collect()]
        for squad in squads:
            endpoint = Endpoint.team_backlog(self._organization, self._project, squad)
            response = self.__get(endpoint)
            if response.status_code == 200 and response.json()["value"]:
                levels = [
                    {item["name"]: item["id"]} for item in response.json()["value"]
                ]
                for item in levels:
                    for level, id_level in item.items():
                        endp_items = Endpoint.backlog_items(
                            self._organization, self._project, squad, id_level
                        )
                        response = self.__get(endp_items).json()
                        if response["workItems"]:
                            id_list = array(
                                "I",
                                [
                                    item["target"]["id"]
                                    for item in response["workItems"]
                                ],
                            )
                            original = self.__get_items(id_list)
                            for item in original:
                                item["Backlog"] = level
                            items.extend(original)
        self._columns["Backlog"] = StructField("Backlog", StringType(), True)
        SCHEMA_ITEMS = StructType([column for column in self._columns.values()])

        data = [tuple(item.values()) for item in items]
        return self.spark.createDataFrame(data, schema=SCHEMA_ITEMS)

    def teams(
        self,
        only: List[str] = None,
        exclude: List[str] = None,
        params_endpoint: str = None,
    ) -> DataFrame:
        """
        `Returns all teams registered in the project.`

        :param only: teams(only = ['Squad 1', 'Squad 2']) -> Only Squad 1 and Squad 2 will be returned.
        :param exclude: teams(exclude = ['Squad 3']) -> Squad 3 will not be returned.
        :param only and exclude:  teams(['Squad 1', 'Squad 2', ...], ['Squad 3']) -> The squads in only will be returned and Squad 3 will not be returned.
        """
        endpoint = Endpoint.teams(self._organization, params_endpoint)
        response = self.__get(endpoint).json()
        clean_response = Process.teams(response)
        teams = (
            [item for item in clean_response if item["Squad"] in only]
            if only
            else clean_response
        )
        remove = (
            [teams.remove(item) for item in teams if item["Squad"] in exclude]
            if exclude
            else None
        )

        data = [tuple(item.values()) for item in teams]
        return self.spark.createDataFrame(data, schema=SCHEMA_ALL_TEAMS)

    def iterations(
        self,
        only: List[str] = None,
        exclude: List[str] = None,
        params_endpoint: str = None,
    ) -> DataFrame:
        """
        `Returns all iterations in the project.`

        :param only: iterations(only = ['iterations 1', 'iterations 2']) -> Only iterations 1 and iterations 2 will be returned.
        :param exclude: iterations(exclude = ['iterations 3']) -> iterations 3 will not be returned.
        :param only and exclude:  iterations(['iterations 1', 'iterations 2', ...], ['iterations 3']) -> The iterations in only will be returned and iterations 3 will not be returned.
        """
        teams = (
            [item["Squad"] for item in self.teams().collect()] if not only else only
        )
        remove = [teams.remove(squad) for squad in exclude] if exclude else None
        iteration_matrix = []
        for squad in teams:
            endpoint = Endpoint.team_iterations(
                self._organization, self._project, squad, params_endpoint
            )
            response = self.__get(endpoint)
            if response.status_code == 200:
                clean_response = Process.team_interation(response.json(), squad)
                iteration_matrix.extend(clean_response)
            else:
                print(
                    f"{squad}: does not exist, or you do not have permission to access it."
                )

        data = [tuple(item.values()) for item in iteration_matrix]
        return self.spark.createDataFrame(data, schema=SCHEMA_ITERATIONS)

    def members(
        self,
        only: List[str] = None,
        exclude: List[str] = None,
        params_endpoint: str = None,
    ) -> DataFrame:
        """
        `Returns all members in the project.`

        :param only: members(only = ['member 1', 'member 2']) -> Only member 1 and member 2 will be returned.
        :param exclude: members(exclude = ['member 3']) -> member 3 will not be returned.
        :param only and exclude:  members(['member 1', 'member 2', ...], ['member 3']) -> The members in only will be returned and member 3 will not be returned.
        """
        teams = (
            self.teams().collect()
            if not only
            else [{"Squad": item} for item in only]
        )
        remove = (
            [teams.remove(item) for item in teams if item["Squad"] in exclude]
            if exclude
            else None
        )
        members_matrix = []

        for squad in teams:
            url = Endpoint.members(
                self._organization, self._project, squad["Squad"], params_endpoint
            )
            response = self.__get(url)
            if response.status_code == 200:
                clear_response = Process.members(response.json(), squad["Squad"])
                members_matrix.append(clear_response)

        members = [member for squad in members_matrix for member in squad]

        data = [tuple(item.values()) for item in members]
        return self.spark.createDataFrame(data, schema=SCHEMA_ALL_MEMBERS)

    def items(self, query: str = None, params_endpoint: str = None) -> DataFrame:
        """
            `Returns all work items in the project. It is possible to filter by SQL in the query parameter set to None`.
            :param query: SQL statements after FROM WorkItems. Ex: [System.WorkItemType] = 'Task' AND [System.AssignedTo] = 'Guilherme Silva'\
                 \n Returns all tasks associated with Guilherme Silva.
        """
        matrix = []
        start = 0
        stop = 19999
        while True:
            sql = f"Select [System.Id] FROM WorkItems WHERE [System.Id] >= {start} AND [System.Id] <= {stop}"
            data = {"query": f"{sql} AND {query}"} if query else {"query": sql}
            endpoint = Endpoint.wiql(self._organization, self._project, params_endpoint)
            response = post(endpoint, headers=self.headers, json=data).json()
            if response["workItems"]:
                id_list = array(
                    "I", sorted({item["id"] for item in response["workItems"]})
                )
                items = self.__get_items(id_list)
                matrix.extend(items)
                start = stop + 1
                stop += 19999
            else:
                break

        SCHEMA_ITEMS = StructType([column for column in self._columns.values()])
        data = [tuple(item.values()) for item in matrix]
        return self.spark.createDataFrame(data, schema=SCHEMA_ITEMS)

    def tags(self) -> DataFrame:
        """
        `Returns all tags registered in the project.`
        """
        endpoint = Endpoint.tags(self._organization, self._project)
        response = self.__get(endpoint).json()
        clear_response = Process.tags(response)
        data = [tuple(item.values()) for item in clear_response]
        return self.spark.createDataFrame(data, schema=SCHEMA_TAGS)

    def __mapping_columns(self) -> bool:
        """
        `Maps all columns in the project.`
        """
        try:
            all_process = Endpoint.all_process(self._organization)
            resp_all_process = self.__get(all_process)
            clean_all_process = Process.all_process(
                resp_all_process.json()
            )  # [{'id':..., 'name':...}]
            for process in clean_all_process:
                project_ref = Endpoint.project_reference(
                    self._organization, process["id"]
                )
                resp_project = self.__get(project_ref).json()
                if "projects" in resp_project:
                    projects = resp_project["projects"]
                    for project in projects:
                        if project["name"].replace(" ", "%20") == self._project:
                            self._process = process["id"]
                            break
            list_work_items = Endpoint.list_work_item_types(
                self._organization, self._process
            )
            resp_list = self.__get(list_work_items)
            clean_all_work_items = Process.all_work_items(resp_list.json())
            for type_id in clean_all_work_items:
                fields = Endpoint.fields_process(
                    self._organization, self._process, type_id
                )
                resp_fields = self.__get(fields)
                schema = Process.clean_fields(resp_fields.json())
                self._columns.update(Process.spark_equivalent(schema))
            print("Ok")
            return True
        except Exception as e:
            raise (f"Error in __mapping_columns: {e}")

    def __get_items(self, items: array[int]) -> List[dict]:
        """
        `Get a list of ids and returns a maximum of 200 work items per request.`
        :param items: An array of unsigned int with all items of the project -> ['85','86','87',...].
        """
        matrix = []
        while True:
            if len(items) >= 200:
                slice_items = items[:200]
                del items[:200]
                matrix.append(self.__max_items(slice_items))
            else:
                slice_items = items
                matrix.append(self.__max_items(slice_items))
                break

        return [item for data in matrix for item in data]

    def __max_items(self, work_items: array[int]) -> List[dict]:
        """
        `Get assignment of `maximum 200 ids` of work items and returns the corresponding data.`
        :param work_items: An array of unsigned int with a maximum size of 200 ids -> [1, 2, 3,..., 4].
        """
        data = []
        id_list = ",".join([str(item) for item in work_items])
        fields = ",".join(self._columns.keys())
        url = Endpoint.work_items(self._organization, self._project, id_list, fields)
        response = self.__get(url).json()
        if "value" in response:
            len_fields = len(response["value"])
            for i in range(0, len_fields):
                fields = response["value"][i]["fields"]
                clean_response = Process.columns(
                    fields, self._columns, self._fields_displayname
                )
                data.append(clean_response)
        return data

    def __get(self, url: str):
        return get(url=url, headers=self.headers)

    def __validade_status(self) -> bool:
        """
        `Validates access credentials with the platform.`
        """
        endpoint = Endpoint.build(self._organization, self._project)
        status = self.__get(endpoint).status_code
        http_status = {
            "302": f"Found.",
            "400": f"Bad Request.",
            "404": f"Not Found.",
            "203": f"No Authorized.",
        }
        if status == 200:
            print("200 - Ok")
            print("Mapping columns...")
            return self.__mapping_columns()
        else:
            err = f"{status} - {http_status.get(str(status))}"
            raise Exception(err)



class Endpoint:
    @classmethod
    def team_iterations(
        cls, organization: str, project: str, squad: str, params: str = None
    ) -> str:
        """
        Get a team's iterations.
        https://docs.microsoft.com/en-us/rest/api/azure/devops/work/iterations/list?view=azure-devops-rest-7.1&tabs=HTTP
        """
        if params:
            return f"https://dev.azure.com/{organization}/{project}/{squad}/_apis/work/teamsettings/iterations?{params}/"
        return f"https://dev.azure.com/{organization}/{project}/{squad}/_apis/work/teamsettings/iterations/"

    @classmethod
    def team_backlog(cls, organization: str, project: str, squad: str) -> str:
        """
        List all backlog levels.
        https://docs.microsoft.com/en-us/rest/api/azure/devops/work/backlogs/list?view=azure-devops-rest-7.1&tabs=HTTP
        """
        return f"https://dev.azure.com/{organization}/{project}/{squad}/_apis/work/backlogs?api-version=7.1-preview.1"

    @classmethod
    def backlog_items(
        cls, organization: str, project: str, squad: str, level: str
    ) -> str:
        """
        Get a list of work items within a backlog level.
        https://docs.microsoft.com/en-us/rest/api/azure/devops/work/backlogs/get-backlog-level-work-items?view=azure-devops-rest-7.1&tabs=HTTP
        """
        return f"https://dev.azure.com/{organization}/{project}/{squad}/_apis/work/backlogs/{level}/workItems?api-version=7.1-preview.1"

    @classmethod
    def work_item_in_iteration(
        cls, organization: str, project: str, squad: str, iteration_id: str
    ) -> str:
        """
        Get the `IDs` of all work items within the iteration(sprint) of a squad.
        https://docs.microsoft.com/en-us/rest/api/azure/devops/work/iterations/get-iteration-work-items?view=azure-devops-rest-7.1&tabs=HTTP

        """
        return f"https://dev.azure.com/{organization}/{project}/{squad}/_apis/work/teamsettings/iterations/{iteration_id}/workitems/"

    @classmethod
    def teams(cls, organization: str, params: str = None) -> str:
        """
        Get a list of all teams.
        https://docs.microsoft.com/en-us/rest/api/azure/devops/core/teams/get-all-teams?view=azure-devops-rest-7.1&tabs=HTTP
        """
        if params:
            return f"https://dev.azure.com/{organization}/_apis/teams?{params}/"
        return f"https://dev.azure.com/{organization}/_apis/teams/"

    @classmethod
    def members(
        cls, organization: str, project: str, squad: str, params: str = None
    ) -> str:
        """
        Get a list of members for a specific team.
        https://docs.microsoft.com/en-us/rest/api/azure/devops/core/teams/get-team-members-with-extended-properties?view=azure-devops-rest-7.1&tabs=HTTP
        """
        if params:
            return f"https://dev.azure.com/{organization}/_apis/projects/{project}/teams/{squad}/members?{params}/"
        return f"https://dev.azure.com/{organization}/_apis/projects/{project}/teams/{squad}/members/"

    @classmethod
    def work_items(cls, organization: str, project: str, id: str, fields: str) -> str:
        """
        Returns a work item by its id. `Can receive a string with a maximum of 200 ids`.
        https://docs.microsoft.com/en-us/rest/api/azure/devops/wit/work-items/get-work-item?view=azure-devops-rest-7.1&tabs=HTTP
        """
        return f"https://dev.azure.com/{organization}/{project}/_apis/wit/workitems?ids={id}&fields={fields}&api-version=6.0"

    @classmethod
    def wiql(cls, organization: str, project: str, params: str = None) -> str:
        """
        Returns all work items for a project.
        https://docs.microsoft.com/en-us/rest/api/azure/devops/wit/wiql/query-by-wiql?view=azure-devops-rest-7.1&tabs=HTTP
        """
        if params:
            return f"https://dev.azure.com/{organization}/{project}/_apis/wit/wiql?{params}&api-version=7.1-preview.2"
        return f"https://dev.azure.com/{organization}/{project}/_apis/wit/wiql?api-version=7.1-preview.2"

    @classmethod
    def tags(cls, organization: str, project: str) -> str:
        """
        Returns all tags for a project.
        https://docs.microsoft.com/en-us/rest/api/azure/devops/wit/tags/list?view=azure-devops-rest-7.1&tabs=HTTP
        """
        return f"https://dev.azure.com/{organization}/{project}/_apis/wit/tags"

    @classmethod
    def build(cls, organization: str, project: str) -> str:
        """
        Just to return 200 quickly.
        https://docs.microsoft.com/en-us/rest/api/azure/devops/build/builds/list?view=azure-devops-rest-7.1
        """
        return f"https://dev.azure.com/{organization}/{project}/_apis/build/builds/"

    @classmethod
    def all_process(cls, organization: str) -> str:
        """
        Returns all processes in an organization.

        """
        return f"https://dev.azure.com/{organization}/_apis/work/processes/"

    @classmethod
    def list_work_item_types(cls, organization: str, process_id: str) -> str:
        """
        Returns all work item types associated with a process.
        https://docs.microsoft.com/en-us/rest/api/azure/devops/processes/work-item-types/list?view=azure-devops-rest-7.1&tabs=HTTP
        """
        return f"https://dev.azure.com/{organization}/_apis/work/processdefinitions/{process_id}/workItemTypes/"

    @classmethod
    def project_reference(cls, organization: str, process_id: str) -> str:
        """
        Returns all projects associated with a process.
        https://docs.microsoft.com/en-us/rest/api/azure/devops/processes/processes/get?view=azure-devops-rest-7.1&tabs=HTTP
        """
        return f"https://dev.azure.com/{organization}/_apis/work/processes/{process_id}?$expand=projects"

    @classmethod
    def fields_process(cls, organization: str, process_id: str, type_id: str) -> str:
        """
        Returns all fields associated with a process.
        """
        return f"https://dev.azure.com/{organization}/_apis/work/processdefinitions/{process_id}/workItemTypes/{type_id}/fields/"



from pyspark.sql.types import DateType, StructType, StringType, StructField


SCHEMA_ITERATIONS = StructType(
    [
        StructField("Iteration_Id", StringType(), True),
        StructField("Iteration_Name", StringType(), True),
        StructField("Iteration_Path", StringType(), True),
        StructField("Iteration_Start_Date", DateType(), True),
        StructField("Iteration_End_Date", DateType(), True),
        StructField("Iteration_State", StringType(), True),
        StructField("Squad", StringType(), True),
    ]
)

SCHEMA_ALL_TEAMS = StructType(
    [StructField("Squad", StringType(), True), StructField("Id", StringType(), True)]
)

SCHEMA_ALL_MEMBERS = StructType(
    [
        StructField("Name", StringType(), True),
        StructField("Id", StringType(), True),
        StructField("Squad", StringType(), True),
    ]
)

SCHEMA_TAGS = StructType(
    [StructField("Id", StringType(), True), StructField("Name", StringType(), True)]
)



from setuptools import setup


text = """
## Azure Devops PySpark:  A productive library to extract data from Azure Devops and apply agile metrics.

## What is it?

Azure Devops PySpark is a Python package that provides the most productive way to extract data from Azure Devops and build agile metrics.
It runs on PySpark, enabling all the features the technology makes available.

## Main Features

- Get authenticated quickly and simply.

- All columns of the project are automatically mapped, use the filter_columns method to filter only those you want to form your dataframes with.
- SparkSession already created in spark variable.
- Get all your organization's backlogs with the method **backlog**.
- Get all your organization's items with the method **items**.
- Get all your organization's iterations with the method **iterations**.
- Get all your organization's members with the method **members**.
- Get all your organization's tags with the method **tags**.
- Get all your organization's teams with the method **teams**.
- Explore the simplicity of Agile class to build powerful metrics for your organization.

## Quick example

```python
from AzureDevopsPySpark import Azure, Agile
from pyspark.sql.functions import datediff #use in agile metrics

devops = Azure('ORGANIZATION', 'PROJECT', 'PERSONAL ACCESS TOKENS')
```

```python
## Filter columns
devops.filter_columns([
    'IterationPath', 'Id', 'State', 'WorkItemType',
    'CreatedDate', 'ClosedDate', 'Iteration_Start_Date', 'Iteration_End_Date'
])


## Pyspark Dataframe
df_backlog = devops.backlog()
df_items = devops.items()
df_iterations = devops.iterations()
df_members = devops.members()
df_tags = devops.tags()
df_teams = devops.teams()
```

```python
## Agile Metrics
agile = Agile()

## A new dataframe
df_agil = df.items.join(df_iterations, 'IterationPath')

## Metrics

## Average time between CreatedDate and ClosedDate of items in the last 90 days.
lead_time = agile.avg(
    df=df_agil,
    ref=[datediff, 'ClosedDate', 'CreatedDate'], # The day difference between the CreatedDate and ClosedDate of each item.
    iteration_path='IterationPath', # GroupBy.
    new='LeadTimeDays', # New column name.
    literal_filter=['ClosedDate >= 90'], # Filtering items from the last 90 days.
    filters={'WorkItemType': 'Task', 'State': 'Closed'} # Custom filters for metric.
).df
```

## Author

The Azure-Devops-PySpark library was written by Guilherme Silva <<https://www.linkedin.com/in/gusantosdev/>> in 2022.

https://github.com/gusantos1/Azure-Devops-Spark

## License

GNU General Public License v3.0.
"""

setup(
    name = 'azure-devops-pyspark',
    version = '1.0.5',
    author = 'Guilherme Silva dos Santos',
    author_email = 'gusantos.ok@gmail.com',
    packages = ['AzureDevopsPySpark'],
    description = 'A productive library to extract data from Azure Devops and apply agile metrics.',
    long_description= text,
    long_description_content_type='text/markdown',
    license = 'GNU General Public License v3.0',
    keywords = 'Azure AzureDevops WorkItems PySpark Agile',
    install_requires= [
        'certifi>=2021.10.8',
        'charset-normalizer>=2.0.12',
        'idna>=3.3',
        'requests>=2.27.1',
        'urllib3>=1.26.9',
        'python-dateutil>=2.8.2',
        'pandas>=1.0.0',
        'PyArrow>=5.0.0'
    ]
)